/*
 * Copyright 2016 Google, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License")
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.netflix.spinnaker.fiat.permissions

import com.fasterxml.jackson.databind.ObjectMapper
import com.netflix.spinnaker.fiat.config.AccountManagerConfig
import com.netflix.spinnaker.fiat.config.FiatAdminConfig
import com.netflix.spinnaker.fiat.config.FiatRoleConfig
import com.netflix.spinnaker.fiat.model.Authorization
import com.netflix.spinnaker.fiat.model.UserPermission
import com.netflix.spinnaker.fiat.model.resources.Account
import com.netflix.spinnaker.fiat.model.resources.Application
import com.netflix.spinnaker.fiat.model.resources.Permissions
import com.netflix.spinnaker.fiat.model.resources.Role
import com.netflix.spinnaker.fiat.model.resources.ServiceAccount
import com.netflix.spinnaker.fiat.providers.AccessControlledResourcePermissionSource
import com.netflix.spinnaker.fiat.providers.AggregatingResourcePermissionProvider
import com.netflix.spinnaker.fiat.providers.DefaultAccountResourceProvider
import com.netflix.spinnaker.fiat.providers.DefaultApplicationResourceProvider
import com.netflix.spinnaker.fiat.providers.DefaultServiceAccountPredicateProvider
import com.netflix.spinnaker.fiat.providers.DefaultServiceAccountResourceProvider
import com.netflix.spinnaker.fiat.providers.ResourcePermissionProvider
import com.netflix.spinnaker.fiat.providers.ResourceProvider
import com.netflix.spinnaker.fiat.providers.internal.ClouddriverService
import com.netflix.spinnaker.fiat.providers.internal.Front50Service
import com.netflix.spinnaker.fiat.roles.UserRolesProvider
import spock.lang.Shared
import spock.lang.Specification
import spock.lang.Subject

class DefaultPermissionsResolverSpec extends Specification {
  UserRolesProvider userRolesProvider = Mock(UserRolesProvider)

  @Shared
  Role anonymous = new Role("anonymous")

  @Shared
  Account noReqGroupsAcct = new Account().setName("noReqGroups")

  @Shared
  Account reqGroup1Acct = new Account().setName("reqGroup1")
                                       .setRequiredGroupMembership(["group1"] as Set)
  @Shared
  Account reqGroup1and2Acct = new Account().setName("reqGroup1and2")
                                           .setRequiredGroupMembership(["group1", "GrouP2"] as Set) // test case insensitivity.

  @Shared
  Account anonymousRead = new Account().setName("anonymousRead")
                                       .setPermissions(Permissions.factory((Authorization.READ): [anonymous.name] as Set, (Authorization.WRITE): ["otherGroup"] as Set))

  @Shared
  ClouddriverService clouddriverService = Mock(ClouddriverService) {
    getAccounts() >> [noReqGroupsAcct, reqGroup1Acct, reqGroup1and2Acct, anonymousRead]
  }

  @Shared
  ResourcePermissionProvider<Account> accountResourcePermissionProvider =
    new AggregatingResourcePermissionProvider<>([new AccessControlledResourcePermissionSource<>()])

  @Shared
  DefaultAccountResourceProvider accountProvider = new DefaultAccountResourceProvider(clouddriverService, accountResourcePermissionProvider)

  @Shared ServiceAccount group1SvcAcct = new ServiceAccount().setName("group1")
                                                             .setMemberOf(["group1"])
  @Shared ServiceAccount group2SvcAcct = new ServiceAccount().setName("group2")
                                                             .setMemberOf(["group2"])

  @Shared
  Front50Service front50Service = Mock(Front50Service) {
    getAllServiceAccounts() >> [group1SvcAcct, group2SvcAcct]
  }

  @Shared
  FiatRoleConfig fiatRoleConfig = Mock(FiatRoleConfig) {
    isOrMode() >> false
  }

  @Shared
  DefaultServiceAccountResourceProvider serviceAccountProvider = new DefaultServiceAccountResourceProvider(
          front50Service, [new DefaultServiceAccountPredicateProvider(fiatRoleConfig) ]
  )

  @Shared
  ResourceProvider<Application> applicationProvider = Mock(ResourceProvider) {
    getAllRestricted(*_) >> []
  }

  @Shared
  List<ResourceProvider> resourceProviders = [accountProvider,
                                              applicationProvider,
                                              serviceAccountProvider]

  def "should resolve the anonymous user permission, when enabled"() {
    setup:
    @Subject DefaultPermissionsResolver resolver = new DefaultPermissionsResolver(
            userRolesProvider, serviceAccountProvider, resourceProviders, new FiatAdminConfig(), new AccountManagerConfig(), new ObjectMapper())

    when:
    def result = resolver.resolveUnrestrictedUser()

    then:
    1 * userRolesProvider.loadUnrestrictedRoles() >> { return [new Role("anonymous")] }

    def expected = new UserPermission().setId("__unrestricted_user__")
                                       .setAccounts([noReqGroupsAcct, anonymousRead] as Set)
                                       .setRoles([anonymous] as Set)
    result == expected
  }

  def "should resolve a single user's permissions"() {
    setup:
    def testUserId = "testUserId"
    UserRolesProvider userRolesProvider = Mock(UserRolesProvider)
    @Subject DefaultPermissionsResolver resolver = new DefaultPermissionsResolver(
            userRolesProvider, serviceAccountProvider, resourceProviders, new FiatAdminConfig(), new AccountManagerConfig(), new ObjectMapper())

    def role1 = new Role("group1")
    def role2 = new Role("gRoUP2") // to test case insensitivity.

    def testUser = new ExternalUser().setId(testUserId).setExternalRoles([role1])

    when:
    resolver.resolve(null as String)

    then:
    thrown IllegalArgumentException

    when:
    def result = resolver.resolve("unknownUser")

    then:
    1 * userRolesProvider.loadRoles({ u -> u.getId() == "unknownUser" }) >> []
    def expected = new UserPermission().setId("unknownUser")
    result == expected

    when:
    result = resolver.resolve(testUserId)
    expected = new UserPermission().setId(testUserId)

    then:
    1 * userRolesProvider.loadRoles({ u -> u.getId() == testUserId }) >> []
    result == expected

    when:
    result = resolver.resolve(testUserId)

    then:
    1 * userRolesProvider.loadRoles({ u -> u.getId() == testUserId }) >> [role2]
    expected.setAccounts([reqGroup1and2Acct] as Set)
            .setServiceAccounts([group2SvcAcct] as Set)
            .setRoles([role2] as Set)
    result == expected

    when: "merge externally provided roles"
    result = resolver.resolveAndMerge(testUser)

    then:
    1 * userRolesProvider.loadRoles({ u -> u.getId() == testUserId }) >> [role2]
    expected.setAccounts([reqGroup1Acct, reqGroup1and2Acct] as Set)
            .setServiceAccounts([group1SvcAcct, group2SvcAcct] as Set)
            .setRoles([role1, role2] as Set)
    result == expected
  }

  def "user should be admin"() {
    FiatAdminConfig fiatAdminConfig = new FiatAdminConfig()
    fiatAdminConfig.getAdmin().getRoles().add("delivery-team")

    def testUserId = "testUserId"
    UserRolesProvider userRolesProvider = Mock(UserRolesProvider)

    @Subject DefaultPermissionsResolver resolver = new DefaultPermissionsResolver(
            userRolesProvider, serviceAccountProvider, resourceProviders, fiatAdminConfig, new AccountManagerConfig(), new ObjectMapper())

    def role1 = new Role("delivery-team")
    def testUser = new ExternalUser().setId(testUserId).setExternalRoles([role1])

    when:
    def result = resolver.resolveAndMerge(testUser)

    then:
    1 * userRolesProvider.loadRoles({ u -> u.getId() == testUserId }) >> [role1]
    def expected = new UserPermission().setId("testUserId")
    expected.setRoles([role1] as Set).setAdmin(true)
            .setServiceAccounts([group1SvcAcct, group2SvcAcct] as Set)
            .setAccounts([reqGroup1Acct, reqGroup1and2Acct, anonymousRead] as Set)
    result == expected

  }

  def "should resolve all user's permissions"() {
    setup:
    UserRolesProvider userRolesProvider = Mock(UserRolesProvider)
    @Subject DefaultPermissionsResolver resolver = new DefaultPermissionsResolver(
            userRolesProvider, serviceAccountProvider, resourceProviders, new FiatAdminConfig(), new AccountManagerConfig(), new ObjectMapper())

    def role1 = new Role("group1")
    def role2 = new Role("group2")

    def extUser1 = new ExternalUser().setId("user1")
    def extUser2 = new ExternalUser().setId("user2")

    when:
    resolver.resolve(null as Collection)

    then:
    thrown IllegalArgumentException

    when:
    1 * userRolesProvider.multiLoadRoles(_) >> [
        user1: [role1],
        user2: [role2],
    ]
    def result = resolver.resolve([extUser1, extUser2])

    then:
    def user1 = new UserPermission().setId("user1")
                                    .setAccounts([reqGroup1Acct, reqGroup1and2Acct] as Set)
                                    .setServiceAccounts([group1SvcAcct] as Set)
                                    .setRoles([role1] as Set)
    def user2 = new UserPermission().setId("user2")
                                    .setAccounts([reqGroup1and2Acct] as Set)
                                    .setServiceAccounts([group2SvcAcct] as Set)
                                    .setRoles([role2] as Set)
    result.remove("user1") == user1
    result.remove("user2") == user2
    result.isEmpty() // Confirm no other values present

    when:
    def extRole = new Role("extRole").setSource(Role.Source.EXTERNAL)
    def extUser3 = new ExternalUser().setId("user3").setExternalRoles([extRole])
    1 * userRolesProvider.multiLoadRoles(_) >> [
        "user3": [role1]
    ]
    result = resolver.resolve([extUser3])

    then:
    def user3 = new UserPermission().setId("user3")
                                    .setAccounts([reqGroup1Acct, reqGroup1and2Acct] as Set)
                                    .setServiceAccounts([group1SvcAcct] as Set)
                                    .setRoles([role1, extRole] as Set)
    result == ["user3": user3]
  }

  def "should resolve service account permissions"() {
    setup:
    UserRolesProvider userRolesProvider = Mock(UserRolesProvider)
    @Subject DefaultPermissionsResolver resolver = new DefaultPermissionsResolver(
            userRolesProvider, serviceAccountProvider, resourceProviders, new FiatAdminConfig(), new AccountManagerConfig(), new ObjectMapper())

    def role1 = new Role(group1SvcAcct.memberOf[0])
    def svc1 = new ExternalUser().setId(group1SvcAcct.name).setExternalRoles([role1])
    def extUser1 = new ExternalUser().setId("user1")

    when:
    0 * userRolesProvider.multiLoadRoles(Collections.emptyList())
    def result = resolver.resolve([svc1])

    then:
    def expectedServiceAcct  = new UserPermission().setId("group1")
            .setAccounts([reqGroup1Acct, reqGroup1and2Acct] as Set)
            .setServiceAccounts([group1SvcAcct] as Set)
            .setRoles([role1] as Set)

    result.remove("group1") == expectedServiceAcct
    result.isEmpty()

    when:
    1 * userRolesProvider.multiLoadRoles([extUser1] as Collection) >> [
            user1: [role1]
    ]
    result = resolver.resolve([svc1, extUser1])

    then:
    def expectedUser1 = new UserPermission().setId("user1")
            .setAccounts([reqGroup1Acct, reqGroup1and2Acct] as Set)
            .setServiceAccounts([group1SvcAcct] as Set)
            .setRoles([role1] as Set)

    result.remove("group1") == expectedServiceAcct
    result.remove("user1") == expectedUser1
    result.isEmpty()
  }

  def "should resolve account manager permissions"() {
    setup:
    def accountManagerConfig = new AccountManagerConfig()
    accountManagerConfig.roles = ['sre']

    def testUsername = 'ron'
    def userRolesProvider = Mock(UserRolesProvider)
    @Subject def resolver = new DefaultPermissionsResolver(userRolesProvider, serviceAccountProvider, resourceProviders, new FiatAdminConfig(), accountManagerConfig, new ObjectMapper())
    def role = new Role('sre')
    def user = new ExternalUser().setId(testUsername).setExternalRoles([role])

    when:
    def result = resolver.resolveAndMerge(user)

    then:
    1 * userRolesProvider.loadRoles({ u -> u.id == testUsername }) >> [role]
    def expected = new UserPermission().setId(testUsername).setRoles(Set.of(role)).setAccountManager(true)
    result == expected
  }

  def "should resolve resources for users and service accounts"() {
    setup:
    FiatAdminConfig fiatAdminConfig = new FiatAdminConfig()
    fiatAdminConfig.getAdmin().getRoles().add("admin-role")

    def role1 = new Role('role1')
    def role2 = new Role('role2')
    def roleAdmin = new Role('admin-role')

    def testApp1 = new Application().setName("app1")
    def testApp2 = new Application().setName("app2")
    def testAppAdmin = new Application().setName("adminOnlyApp")
    ResourceProvider<Application> applicationProvider = Mock(DefaultApplicationResourceProvider)

    @Subject def resolver = new DefaultPermissionsResolver(userRolesProvider, serviceAccountProvider, [applicationProvider], fiatAdminConfig, new AccountManagerConfig(), new ObjectMapper())

    def userToRoles = [
            "user1": [role1],
            "user2": [role1, role2],
            "user3": [role1, roleAdmin],
            "abc@managed-service-accounts": [role1],
    ]
    final Set<String> userRoleSet1 = new HashSet()
    userRoleSet1.add(role1)
    final Set<String> userRoleSet2 = new HashSet()
    userRoleSet2.add(role1)
    userRoleSet2.add(role2)
    final Set<String> userRoleAdmin = new HashSet()
    userRoleAdmin.add(role1)
    userRoleAdmin.add(roleAdmin)
    when:
    applicationProvider.getAllRestricted("user1", userRoleSet1, false) >> [testApp1].toSet()
    applicationProvider.getAllRestricted("user2", userRoleSet2, false) >> [testApp2].toSet()
    applicationProvider.getAllRestricted("user3", userRoleAdmin, true) >> [testAppAdmin].toSet()
    applicationProvider.getAllRestricted("abc@managed-service-accounts", userRoleSet1, false) >> [testApp1].toSet()
    def result = resolver.resolveResources(userToRoles)

    then:
    result == [
            "user1": new UserPermission()
                    .setId("user1")
                    .setRoles([role1] as Set<Role>)
                    .setAdmin(false)
                    .addResources([testApp1]),
            "user2": new UserPermission()
                    .setId("user2")
                    .setRoles([role1, role2] as Set<Role>)
                    .setAdmin(false)
                    .addResources([testApp2]),
            "user3": new UserPermission()
                    .setId("user3")
                    .setRoles([role1, roleAdmin] as Set<Role>)
                    .setAdmin(true)
                    .addResources([testAppAdmin]),
            "abc@managed-service-accounts": new UserPermission()
                    .setId("abc@managed-service-accounts")
                    .setRoles([role1] as Set<Role>)
                    .setAdmin(false)
                    .addResources([testApp1])
    ]
  }
}
